@@ -16,11 +16,12 @@ import os, pickle
 import build
 import dependencies
 import mesonlib
+import json
 from coredata import MesonException
 
 class TestSerialisation:
     def __init__(self, name, fname, is_cross, exe_wrapper, is_parallel, cmd_args, env,
-                 should_fail, valgrind_args):
+                 should_fail, valgrind_args, timeout, extra_paths):
         self.name = name
         self.fname = fname
         self.is_cross = is_cross
@@ -30,6 +31,8 @@ class TestSerialisation:
         self.env = env
         self.should_fail = should_fail
         self.valgrind_args = valgrind_args
+        self.timeout = timeout
+        self.extra_paths = extra_paths
 
 # This class contains the basic functionality that is needed by all backends.
 # Feel free to move stuff in and out of it as you see fit.
@@ -200,10 +203,10 @@ class Backend():
     def generate_basic_compiler_args(self, target, compiler):
         commands = []
         commands += compiler.get_always_args()
+        if self.environment.coredata.buildtype != 'plain':
+            commands += compiler.get_warn_args(self.environment.coredata.warning_level)
         commands += self.build.get_global_args(compiler)
         commands += self.environment.coredata.external_args[compiler.get_language()]
-        if self.environment.coredata.buildtype != 'plain':
-            commands += compiler.get_std_warn_args()
         commands += target.get_extra_args(compiler.get_language())
         commands += compiler.get_buildtype_args(self.environment.coredata.buildtype)
         if self.environment.coredata.coverage:
@@ -217,6 +220,12 @@ class Backend():
             if isinstance(target, build.Executable):
                 commands += dep.get_exe_args()
 
+        # Fortran rquires extra include directives.
+        if compiler.language == 'fortran':
+            for lt in target.link_targets:
+                priv_dir = os.path.join(lt.subdir, lt.get_basename() + lt.type_suffix())
+                incflag = compiler.get_include_args(priv_dir)
+                commands += incflag
         return commands
 
     def build_target_link_arguments(self, compiler, deps):
@@ -237,6 +246,17 @@ class Backend():
             args += self.build_target_link_arguments(compiler, d.get_dependencies())
         return args
 
+    def determine_windows_extra_paths(self, target):
+        '''On Windows there is no such thing as an rpath.
+        We must determine all locations of DLLs that this exe
+        links to and return them so they can be used in unit
+        tests.'''
+        if not isinstance(target, build.Executable):
+            print(target)
+            return []
+        prospectives = target.get_transitive_rpaths()
+        return [os.path.join(self.environment.get_build_dir(), i) for i in prospectives if len(i) > 0]
+
     def write_test_file(self, datafile):
         arr = []
         for t in self.build.get_tests():
@@ -245,13 +265,23 @@ class Backend():
                 fname = exe.fullpath
             else:
                 fname = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(t.get_exe()))]
-            is_cross = self.environment.is_cross_build()
+            is_cross = self.environment.is_cross_build() and self.environment.cross_info.need_cross_compiler()
             if is_cross:
-                exe_wrapper = self.environment.cross_info.get('exe_wrapper', None)
+                exe_wrapper = self.environment.cross_info.config['binaries'].get('exe_wrapper', None)
             else:
                 exe_wrapper = None
+            if mesonlib.is_windows():
+                extra_paths = self.determine_windows_extra_paths(exe)
+            else:
+                extra_paths = []
+            cmd_args = []
+            for a in t.cmd_args:
+                if isinstance(a, mesonlib.File):
+                    a = os.path.join(self.environment.get_build_dir(), a.rel_to_builddir(self.build_to_src))
+                cmd_args.append(a)
             ts = TestSerialisation(t.get_name(), fname, is_cross, exe_wrapper,
-                                   t.is_parallel, t.cmd_args, t.env, t.should_fail, t.valgrind_args)
+                                   t.is_parallel, cmd_args, t.env, t.should_fail, t.valgrind_args,
+                                   t.timeout, extra_paths)
             arr.append(ts)
         pickle.dump(arr, datafile)
 
@@ -280,3 +310,14 @@ class Backend():
                 ofile.write(' ')
             ofile.write('\n')
 
+
+    def generate_depmf_install(self, d):
+        if self.build.dep_manifest_name is None:
+            return
+        ifilename = os.path.join(self.environment.get_build_dir(), 'depmf.json')
+        ofilename = os.path.join(self.environment.get_prefix(), self.build.dep_manifest_name)
+        mfobj = {'type': 'dependency manifest',
+                 'version': '1.0'}
+        mfobj['projects'] = self.build.dep_manifest
+        open(ifilename, 'w').write(json.dumps(mfobj))
+        d.data.append([ifilename, ofilename])
