@@ -57,6 +57,7 @@ class Build:
 
     def __init__(self, environment):
         self.project_name = 'name of master project'
+        self.project_version = None
         self.environment = environment
         self.projects = {}
         self.targets = {}
@@ -72,8 +73,10 @@ class Build:
         self.pot = []
         self.subprojects = {}
         self.pkgconfig_gens = []
-        self.install_script = None
+        self.install_scripts = []
         self.install_dirs = []
+        self.dep_manifest_name = None
+        self.dep_manifest = {}
 
     def has_language(self, language):
         for i in self.compilers:
@@ -139,9 +142,10 @@ class ExtractedObjects():
         self.srclist = srclist
 
 class BuildTarget():
-    def __init__(self, name, subdir, is_cross, sources, objects, environment, kwargs):
+    def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         self.name = name
         self.subdir = subdir
+        self.subproject = subproject # Can not be calculated from subdir as subproject dirname can be changed per project.
         self.is_cross = is_cross
         self.sources = []
         self.objects = []
@@ -159,10 +163,15 @@ class BuildTarget():
         self.process_objectlist(objects)
         self.process_kwargs(kwargs, environment)
         self.check_unknown_kwargs(kwargs)
-        if len(self.sources) == 0 and len(self.generated) == 0:
+        if len(self.sources) == 0 and \
+            len(self.generated) == 0 and \
+            len(self.objects) == 0:
             raise InvalidArguments('Build target %s has no sources.' % name)
         self.validate_sources()
 
+    def get_id(self):
+        return self.subproject + ':' + self.name + self.type_suffix()
+
     def check_unknown_kwargs(self, kwargs):
         # Override this method in derived classes that have more
         # keywords.
@@ -255,6 +264,9 @@ class BuildTarget():
                 obj_src.append(src)
         return ExtractedObjects(self, obj_src)
 
+    def extract_all_objects(self):
+        return ExtractedObjects(self, self.sources)
+
     def get_rpaths(self):
         return self.get_transitive_rpaths()
 
@@ -573,8 +585,8 @@ class GeneratedList():
         return self.generator
 
 class Executable(BuildTarget):
-    def __init__(self, name, subdir, is_cross, sources, objects, environment, kwargs):
-        super().__init__(name, subdir, is_cross, sources, objects, environment, kwargs)
+    def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
+        super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
         self.prefix = ''
         self.suffix = environment.get_exe_suffix()
         suffix = environment.get_exe_suffix()
@@ -589,8 +601,8 @@ class Executable(BuildTarget):
         return "@exe"
 
 class StaticLibrary(BuildTarget):
-    def __init__(self, name, subdir, is_cross, sources, objects, environment, kwargs):
-        super().__init__(name, subdir, is_cross, sources, objects, environment, kwargs)
+    def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
+        super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
         if len(self.sources) > 0 and self.sources[0].endswith('.cs'):
             raise InvalidArguments('Static libraries not supported for C#.')
         self.prefix = environment.get_static_lib_prefix()
@@ -607,10 +619,10 @@ class StaticLibrary(BuildTarget):
         return "@sta"
 
 class SharedLibrary(BuildTarget):
-    def __init__(self, name, subdir, is_cross, sources, objects, environment, kwargs):
+    def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         self.version = None
         self.soversion = None
-        super().__init__(name, subdir, is_cross, sources, objects, environment, kwargs);
+        super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs);
         if len(self.sources) > 0 and self.sources[0].endswith('.cs'):
             self.suffix = 'dll'
             self.prefix = 'lib'
@@ -618,6 +630,7 @@ class SharedLibrary(BuildTarget):
             self.prefix = environment.get_shared_lib_prefix()
             self.suffix = environment.get_shared_lib_suffix()
         self.importsuffix = environment.get_import_lib_suffix()
+        self.filename = self.prefix + self.name + '.' + self.suffix
 
     def process_kwargs(self, kwargs, environment):
         super().process_kwargs(kwargs, environment)
@@ -699,6 +712,9 @@ class CustomTarget:
             mlog.log(mlog.bold('Warning:'), 'Unknown keyword arguments in target %s: %s' %
                      (self.name, ', '.join(unknowns)))
 
+    def get_id(self):
+        return self.name + self.type_suffix()
+
     def process_kwargs(self, kwargs):
         self.sources = kwargs.get('input', [])
         if not isinstance(self.sources, list):
@@ -806,6 +822,9 @@ class RunTarget:
         self.args = args
         self.subdir = subdir
 
+    def get_id(self):
+        return self.name + self.type_suffix()
+
     def get_basename(self):
         return self.name
 
@@ -831,8 +850,8 @@ class RunTarget:
         return "@run"
 
 class Jar(BuildTarget):
-    def __init__(self, name, subdir, is_cross, sources, objects, environment, kwargs):
-        super().__init__(name, subdir, is_cross, sources, objects, environment, kwargs);
+    def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
+        super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs);
         for s in self.sources:
             if not s.endswith('.java'):
                 raise InvalidArguments('Jar source %s is not a java file.' % s)
@@ -889,3 +908,8 @@ class PkgConfigGenerator():
         self.description = description
         self.version = version
         self.filebase = filebase
+
+class InstallScript:
+    def __init__(self, cmd_arr):
+        assert(isinstance(cmd_arr, list))
+        self.cmd_arr = cmd_arr
