@@ -302,32 +302,55 @@ class GeneratedListHolder(InterpreterObject):
     def add_file(self, a):
         self.held_object.add_file(a)
 
-class Build(InterpreterObject):
+class BuildMachine(InterpreterObject):
     def __init__(self):
         InterpreterObject.__init__(self)
-        self.methods.update({'name' : self.get_name_method,
+        self.methods.update({'system' : self.system_method,
+                             'cpu' : self.cpu_method,
+                             'endian' : self.endian_method,
                             })
 
-    def get_name_method(self, args, kwargs):
+    # Python is inconsistent in its platform module.
+    # It returns different values for the same cpu.
+    # For x86 it might return 'x86', 'i686' or somesuch.
+    # Do some canonicalization.
+    def cpu_method(self, args, kwargs):
+        trial = platform.machine().lower()
+        if trial.startswith('i') and trial.endswith('86'):
+            return 'x86'
+        # This might be wrong. Maybe we should return the more
+        # specific string such as 'armv7l'. Need to get user
+        # feedback first.
+        if trial.startswith('arm'):
+            return 'arm'
+        # Add fixes here as bugs are reported.
+        return trial
+
+    def system_method(self, args, kwargs):
         return platform.system().lower()
 
-# This currently returns data for the current environment.
-# It should return info for the target host.
-class Host(InterpreterObject):
-    def __init__(self, envir):
+    def endian_method(self, args, kwargs):
+        return sys.byteorder
+
+# This class will provide both host_machine and
+# target_machine
+class CrossMachineInfo(InterpreterObject):
+    def __init__(self, cross_info):
         InterpreterObject.__init__(self)
-        self.environment = envir
-        self.methods.update({'name' : self.get_name_method,
-                             'is_big_endian' : self.is_big_endian_method,
+        self.info = cross_info
+        self.methods.update({'system' : self.system_method,
+                             'cpu' : self.cpu_method,
+                             'endian' : self.endian_method,
                             })
 
-    def get_name_method(self, args, kwargs):
-        if self.environment.is_cross_build():
-            return self.environment.cross_info.get('name')
-        return platform.system().lower()
+    def system_method(self, args, kwargs):
+        return self.info['system']
 
-    def is_big_endian_method(self, args, kwargs):
-        return sys.byteorder != 'little'
+    def cpu_method(self, args, kwargs):
+        return self.info['cpu']
+
+    def endian_method(self, args, kwargs):
+        return self.info['endian']
 
 class IncludeDirsHolder(InterpreterObject):
     def __init__(self, curdir, dirs):
@@ -429,7 +452,8 @@ class BuildTargetHolder(InterpreterObject):
     def __init__(self, target):
         super().__init__()
         self.held_object = target
-        self.methods.update({'extract_objects' : self.extract_objects_method})
+        self.methods.update({'extract_objects' : self.extract_objects_method,
+                             'extract_all_objects' : self.extract_all_objects_method})
 
     def is_cross(self):
         return self.held_object.is_cross()
@@ -438,6 +462,10 @@ class BuildTargetHolder(InterpreterObject):
         gobjs = self.held_object.extract_objects(args)
         return GeneratedObjectsHolder(gobjs)
 
+    def extract_all_objects_method(self, args, kwargs):
+        gobjs = self.held_object.extract_all_objects()
+        return GeneratedObjectsHolder(gobjs)
+
 class ExecutableHolder(BuildTargetHolder):
     def __init__(self, target):
         super().__init__(target)
@@ -470,7 +498,7 @@ class RunTargetHolder(InterpreterObject):
         self.held_object = build.RunTarget(name, command, args, subdir)
 
 class Test(InterpreterObject):
-    def __init__(self, name, exe, is_parallel, cmd_args, env, should_fail, valgrind_args):
+    def __init__(self, name, exe, is_parallel, cmd_args, env, should_fail, valgrind_args, timeout):
         InterpreterObject.__init__(self)
         self.name = name
         self.exe = exe
@@ -479,6 +507,7 @@ class Test(InterpreterObject):
         self.env = env
         self.should_fail = should_fail
         self.valgrind_args = valgrind_args
+        self.timeout = timeout
 
     def get_exe(self):
         return self.exe
@@ -670,6 +699,7 @@ class ModuleHolder(InterpreterObject):
         state.subdir = self.interpreter.subdir
         state.environment = self.interpreter.environment
         state.project_name = self.interpreter.build.project_name
+        state.project_version = self.interpreter.build.dep_manifest[self.interpreter.active_projectname]
         state.compilers = self.interpreter.build.compilers
         state.targets = self.interpreter.build.targets
         state.headers = self.interpreter.build.get_headers()
@@ -693,10 +723,12 @@ class MesonMain(InterpreterObject):
                              'current_build_dir' : self.current_build_dir_method,
                              'source_root' : self.source_root_method,
                              'build_root' : self.build_root_method,
-                             'set_install_script' : self.set_install_script_method,
+                             'add_install_script' : self.add_install_script_method,
+                             'install_dependency_manifest': self.install_dependency_manifest_method,
+                             'project_version': self.project_version_method,
                             })
 
-    def set_install_script_method(self, args, kwargs):
+    def add_install_script_method(self, args, kwargs):
         if len(args) != 1:
             raise InterpreterException('Set_install_script takes exactly one argument.')
         check_stringlist(args)
@@ -705,7 +737,7 @@ class MesonMain(InterpreterObject):
                                   self.interpreter.subdir, scriptbase)
         if not os.path.isfile(scriptfile):
             raise InterpreterException('Can not find install script %s.' % scriptbase)
-        self.build.install_script = scriptfile
+        self.build.install_scripts.append(build.InstallScript([scriptfile]))
 
     def current_source_dir_method(self, args, kwargs):
         src = self.interpreter.environment.source_dir
@@ -728,9 +760,9 @@ class MesonMain(InterpreterObject):
         return self.interpreter.environment.build_dir
 
     def has_exe_wrapper_method(self, args, kwargs):
-        if self.is_cross_build_method(None, None):
-            return 'exe_wrap' in  self.build.environment.cross_info
-        return True # This is semantically confusing.
+        if self.is_cross_build_method(None, None) and 'binaries' in self.build.environment.cross_info.config:
+            return 'exe_wrap' in self.build.environment.cross_info.config['binaries']
+        return True  # This is semantically confusing.
 
     def is_cross_build_method(self, args, kwargs):
         return self.build.environment.is_cross_build()
@@ -762,6 +794,16 @@ class MesonMain(InterpreterObject):
     def is_subproject_method(self, args, kwargs):
         return self.interpreter.is_subproject()
 
+    def install_dependency_manifest_method(self, args, kwargs):
+        if len(args) != 1:
+            raise InterpreterException('Must specify manifest install file name')
+        if not isinstance(args[0], str):
+            raise InterpreterException('Argument must be a string.')
+        self.build.dep_manifest_name = args[0]
+
+    def project_version_method(self, args, kwargs):
+        return self.build.dep_manifest[self.interpreter.active_projectname]
+
 class Interpreter():
 
     def __init__(self, build, subproject='', subdir='', subproject_dir='subprojects'):
@@ -772,7 +814,7 @@ class Interpreter():
         self.subproject_dir = subproject_dir
         option_file = os.path.join(self.source_root, self.subdir, 'meson_options.txt')
         if os.path.exists(option_file):
-            oi = optinterpreter.OptionInterpreter(self.subproject,\
+            oi = optinterpreter.OptionInterpreter(self.subproject, \
                                                   self.build.environment.cmd_line_options)
             oi.process(option_file)
             self.build.environment.merge_options(oi.options)
@@ -791,8 +833,20 @@ class Interpreter():
         self.sanity_check_ast()
         self.variables = {}
         self.builtin = {}
-        self.builtin['build'] = Build()
-        self.builtin['host'] = Host(build.environment)
+        self.builtin['build_machine'] = BuildMachine()
+        if not self.build.environment.is_cross_build():
+            self.builtin['host_machine'] = self.builtin['build_machine']
+            self.builtin['target_machine'] = self.builtin['build_machine']
+        else:
+            cross_info = self.build.environment.cross_info
+            if cross_info.has_host():
+                self.builtin['host_machine'] = CrossMachineInfo(cross_info.config['host_machine'])
+            else:
+                self.builtin['host_machine'] = self.builtin['build_machine']
+            if cross_info.has_target():
+                self.builtin['target_machine'] = CrossMachineInfo(cross_info.config['target_machine'])
+            else:
+                self.builtin['target_machine'] = self.builtin['host_machine']
         self.builtin['meson'] = MesonMain(build, self)
         self.environment = build.environment
         self.build_func_dict()
@@ -826,6 +880,7 @@ class Interpreter():
                       'configure_file' : self.func_configure_file,
                       'include_directories' : self.func_include_directories,
                       'add_global_arguments' : self.func_add_global_arguments,
+                      'add_languages' : self.func_add_languages,
                       'find_program' : self.func_find_program,
                       'find_library' : self.func_find_library,
                       'configuration_data' : self.func_configuration_data,
@@ -867,6 +922,12 @@ class Interpreter():
                 outvalues.append(self.module_method_callback(v))
             elif isinstance(v, build.GeneratedList):
                 outvalues.append(GeneratedListHolder(v))
+            elif isinstance(v, build.RunTarget):
+                if v.name in self.build.targets:
+                    raise InterpreterException('Tried to create target %s which already exists.' % v.name)
+                self.build.targets[v.name] = v
+            elif isinstance(v, build.InstallScript):
+                self.build.install_scripts.append(v)
             else:
                 print(v)
                 raise InterpreterException('Module returned a value of unknown type.')
@@ -1006,6 +1067,10 @@ class Interpreter():
             return self.evaluate_arithmeticstatement(cur)
         elif isinstance(cur, mparser.ForeachClauseNode):
             return self.evaluate_foreach(cur)
+        elif isinstance(cur, mparser.PlusAssignmentNode):
+            return self.evaluate_plusassign(cur)
+        elif isinstance(cur, mparser.IndexNode):
+            return self.evaluate_indexing(cur)
         elif self.is_elementary_type(cur):
             return cur
         else:
@@ -1103,7 +1168,6 @@ class Interpreter():
             raise InterpreterException('Recursive include of subprojects: %s.' % incpath)
         if dirname in self.subprojects:
             return self.subprojects[dirname]
-        subdir = os.path.join(self.subproject_dir, dirname)
         r = wrap.Resolver(os.path.join(self.build.environment.get_source_dir(), self.subproject_dir))
         resolved = r.resolve(dirname)
         if resolved is None:
@@ -1116,7 +1180,9 @@ class Interpreter():
         subi.subprojects = self.subprojects
 
         subi.subproject_stack = self.subproject_stack + [dirname]
+        current_active = self.active_projectname
         subi.run()
+        self.active_projectname = current_active
         mlog.log('\nSubproject', mlog.bold(dirname), 'finished.')
         self.build.subprojects[dirname] = True
         self.subprojects.update(subi.subprojects)
@@ -1148,13 +1214,13 @@ class Interpreter():
 
     @stringArgs
     def func_project(self, node, args, kwargs):
-        if len(args)< 2:
+        if len(args) < 2:
             raise InvalidArguments('Not enough arguments to project(). Needs at least the project name and one language')
-        if list(kwargs.keys()) != ['subproject_dir'] and len(kwargs) != 0:
-            raise InvalidArguments('project() only accepts the keyword argument "subproject_dir"')
 
         if not self.is_subproject():
             self.build.project_name = args[0]
+        self.active_projectname = args[0]
+        self.build.dep_manifest[args[0]] = kwargs.get('version', 'undefined')
         if self.subproject in self.build.projects:
             raise InvalidCode('Second call to project().')
         if not self.is_subproject() and 'subproject_dir' in kwargs:
@@ -1168,6 +1234,11 @@ class Interpreter():
             if not 'c' in langs:
                 raise InterpreterException('Compiling Vala requires a C compiler')
 
+    @noKwargs
+    @stringArgs
+    def func_add_languages(self, node, args, kwargs):
+        self.add_languages(node, args)
+
     @noKwargs
     def func_message(self, node, args, kwargs):
         # reduce arguments again to avoid flattening posargs
@@ -1177,7 +1248,7 @@ class Interpreter():
 
         arg = posargs[0]
         if isinstance(arg, list):
-            argstr =  stringifyUserArguments(arg)
+            argstr = stringifyUserArguments(arg)
         elif isinstance(arg, str):
             argstr = arg
         elif isinstance(arg, int):
@@ -1195,7 +1266,7 @@ class Interpreter():
         raise InterpreterException('Error encountered: ' + args[0])
 
     def add_languages(self, node, args):
-        is_cross = self.environment.is_cross_build()
+        need_cross_compiler = self.environment.is_cross_build() and self.environment.cross_info.need_cross_compiler()
         for lang in args:
             lang = lang.lower()
             if lang in self.coredata.compilers:
@@ -1205,45 +1276,46 @@ class Interpreter():
                 cross_comp = None
                 if lang == 'c':
                     comp = self.environment.detect_c_compiler(False)
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = self.environment.detect_c_compiler(True)
                 elif lang == 'cpp':
                     comp = self.environment.detect_cpp_compiler(False)
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = self.environment.detect_cpp_compiler(True)
                 elif lang == 'objc':
                     comp = self.environment.detect_objc_compiler(False)
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = self.environment.detect_objc_compiler(True)
                 elif lang == 'objcpp':
                     comp = self.environment.detect_objcpp_compiler(False)
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = self.environment.detect_objcpp_compiler(True)
                 elif lang == 'java':
                     comp = self.environment.detect_java_compiler()
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = comp # Java is platform independent.
                 elif lang == 'cs':
                     comp = self.environment.detect_cs_compiler()
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = comp # C# is platform independent.
                 elif lang == 'vala':
                     comp = self.environment.detect_vala_compiler()
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = comp # Vala is too (I think).
                 elif lang == 'rust':
                     comp = self.environment.detect_rust_compiler()
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = comp # FIXME, probably not correct.
                 elif lang == 'fortran':
                     comp = self.environment.detect_fortran_compiler(False)
-                    if is_cross:
+                    if need_cross_compiler:
                         cross_comp = self.environment.detect_fortran_compiler(True)
                 else:
                     raise InvalidCode('Tried to use unknown language "%s".' % lang)
                 comp.sanity_check(self.environment.get_scratch_dir())
                 self.coredata.compilers[lang] = comp
                 if cross_comp is not None:
+                    cross_comp.sanity_check(self.environment.get_scratch_dir())
                     self.coredata.cross_compilers[lang] = cross_comp
             mlog.log('Native %s compiler: ' % lang, mlog.bold(' '.join(comp.get_exelist())), ' (%s %s)' % (comp.id, comp.version), sep='')
             if not comp.get_language() in self.coredata.external_args:
@@ -1251,9 +1323,11 @@ class Interpreter():
                 self.coredata.external_args[comp.get_language()] = ext_compile_args
                 self.coredata.external_link_args[comp.get_language()] = ext_link_args
             self.build.add_compiler(comp)
-            if is_cross:
+            if need_cross_compiler:
                 mlog.log('Cross %s compiler: ' % lang, mlog.bold(' '.join(cross_comp.get_exelist())), ' (%s %s)' % (cross_comp.id, cross_comp.version), sep='')
                 self.build.add_cross_compiler(cross_comp)
+            if self.environment.is_cross_build() and not need_cross_compiler:
+                self.build.add_cross_compiler(comp)
 
     def func_find_program(self, node, args, kwargs):
         self.validate_arguments(args, 1, [str])
@@ -1377,11 +1451,18 @@ class Interpreter():
         self.add_target(name, tg.held_object)
         return tg
 
-    @stringArgs
     @noKwargs
     def func_run_target(self, node, args, kwargs):
         if len(args) < 2:
             raise InterpreterException('Incorrect number of arguments')
+        for i in args:
+            try:
+                i = i.held_object
+            except AttributeError:
+                pass
+            if not isinstance(i, (str, build.BuildTarget)):
+                mlog.debug('Wrong type:', str(i))
+                raise InterpreterException('Invalid argument to run_target.')
         name = args[0]
         command = args[1]
         cmd_args = args[2:]
@@ -1408,7 +1489,7 @@ class Interpreter():
         if not isinstance(cmd_args, list):
             cmd_args = [cmd_args]
         for i in cmd_args:
-            if not isinstance(i, str):
+            if not isinstance(i, (str, mesonlib.File)):
                 raise InterpreterException('Command line arguments must be strings')
         envlist = kwargs.get('env', [])
         if not isinstance(envlist, list):
@@ -1432,7 +1513,10 @@ class Interpreter():
         should_fail = kwargs.get('should_fail', False)
         if not isinstance(should_fail, bool):
             raise InterpreterException('Keyword argument should_fail must be a boolean.')
-        t = Test(args[0], args[1].held_object, par, cmd_args, env, should_fail, valgrind_args)
+        timeout = kwargs.get('timeout', 30)
+        if not isinstance(timeout, int):
+            raise InterpreterException('Timeout must be an integer.')
+        t = Test(args[0], args[1].held_object, par, cmd_args, env, should_fail, valgrind_args, timeout)
         self.build.tests.append(t)
         mlog.debug('Adding test "', mlog.bold(args[0]), '".', sep='')
 
@@ -1463,7 +1547,7 @@ class Interpreter():
         self.visited_subdirs[subdir] = True
         self.subdir = subdir
         try:
-            os.mkdir(os.path.join(self.environment.build_dir, subdir))
+            os.makedirs(os.path.join(self.environment.build_dir, subdir))
         except FileExistsError:
             pass
         buildfilename = os.path.join(self.subdir, environment.build_filename)
@@ -1489,7 +1573,7 @@ class Interpreter():
 
     @stringArgs
     def func_install_subdir(self, node, args, kwargs):
-        if len(args ) != 1:
+        if len(args) != 1:
             raise InvalidArguments('Install_subdir requires exactly one argument.')
         if not 'install_dir' in kwargs:
             raise InvalidArguments('Missing keyword argument install_dir')
@@ -1519,7 +1603,8 @@ class Interpreter():
                 raise InterpreterException('Argument "configuration" is not of type configuration_data')
 
             conffile = os.path.join(self.subdir, inputfile)
-            self.build_def_files.append(conffile)
+            if conffile not in self.build_def_files:
+                self.build_def_files.append(conffile)
             os.makedirs(os.path.join(self.environment.build_dir, self.subdir), exist_ok=True)
             ifile_abs = os.path.join(self.environment.source_dir, self.subdir, inputfile)
             ofile_abs = os.path.join(self.environment.build_dir, self.subdir, output)
@@ -1534,12 +1619,7 @@ class Interpreter():
             raise InterpreterException('Configure_file must have either "configuration" or "command".')
         if isinstance(kwargs.get('install_dir', None), str):
             self.build.data.append(Data(False, self.subdir, [output], kwargs))
-        # FIXME, HORROR, this is a massive hack to get this working. The correct
-        # solution is to finally do the refactoring where source files are no
-        # longer strings but actual objects. This is a major undertaking
-        # and will only be done after the next release.
-        outputfile = os.path.join(self.environment.build_dir, self.subdir, output)
-        return outputfile
+        return mesonlib.File.from_built_file(self.subdir, output)
 
     @stringArgs
     @noKwargs
@@ -1605,7 +1685,7 @@ class Interpreter():
                                    % name)
         # To permit an executable and a shared library to have the
         # same name, such as "foo.exe" and "libfoo.a".
-        idname = name + tobj.type_suffix()
+        idname = tobj.get_id()
         if idname in self.build.targets:
             raise InvalidCode('Tried to create target "%s", but a target of that name already exists.' % name)
         self.build.targets[idname] = tobj
@@ -1646,7 +1726,7 @@ class Interpreter():
         else:
             mlog.debug('Unknown target type:', str(targetholder))
             raise RuntimeError('Unreachable code')
-        target = targetclass(name, self.subdir, is_cross, sources, objs, self.environment, kwargs)
+        target = targetclass(name, self.subdir, self.subproject, is_cross, sources, objs, self.environment, kwargs)
         l = targetholder(target)
         self.add_target(name, l.held_object)
         self.global_args_frozen = True
@@ -1780,6 +1860,8 @@ class Interpreter():
     def array_method_call(self, obj, method_name, args):
         if method_name == 'contains':
             return self.check_contains(obj, args)
+        elif method_name == 'length':
+            return len(obj)
         elif method_name == 'get':
             index = args[0]
             if not isinstance(index, int):
@@ -1828,6 +1910,35 @@ class Interpreter():
             self.set_variable(varname, item)
             self.evaluate_codeblock(node.block)
 
+    def evaluate_plusassign(self, node):
+        assert(isinstance(node, mparser.PlusAssignmentNode))
+        varname = node.var_name
+        addition = self.evaluate_statement(node.value)
+        # Remember that all variables are immutable. We must always create a
+        # full new variable and then assign it.
+        old_variable = self.get_variable(varname)
+        if not isinstance(old_variable, list):
+            raise InvalidArguments('The += operator currently only works with arrays.')
+        # Add other data types here.
+        else:
+            if isinstance(addition, list):
+                new_value = old_variable + addition
+            else:
+                new_value = old_variable + [addition]
+        self.set_variable(varname, new_value)
+
+    def evaluate_indexing(self, node):
+        assert(isinstance(node, mparser.IndexNode))
+        iobject = self.evaluate_statement(node.iobject)
+        if not isinstance(iobject, list):
+            raise InterpreterException('Tried to index a non-array object.')
+        index = self.evaluate_statement(node.index)
+        if not isinstance(index, int):
+            raise InterpreterException('Index value is not an integer.')
+        if index < -len(iobject) or index >= len(iobject):
+            raise InterpreterException('Index %d out of bounds of array of size %d.' % (index, len(iobject)))
+        return iobject[index]
+
     def is_elementary_type(self, v):
         if isinstance(v, (int, float, str, bool, list)):
             return True
