@@ -16,11 +16,13 @@
 functionality such as gobject-introspection and gresources.'''
 
 import build
-import os
+import os, sys
 import subprocess
 from coredata import MesonException
 import mlog
 
+girwarning_printed = False
+
 class GnomeModule:
 
     def compile_resources(self, state, args, kwargs):
@@ -49,7 +51,14 @@ class GnomeModule:
             girtarget = girtarget.held_object
         if not isinstance(girtarget, (build.Executable, build.SharedLibrary)):
             raise MesonException('Gir target must be an executable or shared library')
-        pkgstr = subprocess.check_output(['pkg-config', '--cflags', 'gobject-introspection-1.0'])
+        try:
+            pkgstr = subprocess.check_output(['pkg-config', '--cflags', 'gobject-introspection-1.0'])
+        except Exception:
+            global girwarning_printed
+            if not girwarning_printed:
+                mlog.log(mlog.bold('Warning:'), 'gobject-introspection dependency was not found, disabling gir generation.')
+                girwarning_printed = True
+            return []
         pkgargs = pkgstr.decode().strip().split()
         ns = kwargs.pop('namespace')
         nsversion = kwargs.pop('nsversion')
@@ -59,15 +68,22 @@ class GnomeModule:
 
         scan_command = ['g-ir-scanner', '@INPUT@']
         scan_command += pkgargs
-        scan_command += ['--namespace='+ns, '--nsversion=' + nsversion, '--warn-all',
+        scan_command += ['--no-libtool', '--namespace='+ns, '--nsversion=' + nsversion, '--warn-all',
                          '--output', '@OUTPUT@']
 
+        extra_args = kwargs.pop('extra_args', [])
+        if not isinstance(extra_args, list):
+            extra_args = [extra_args]
+        scan_command += extra_args
+
         for incdirs in girtarget.include_dirs:
             for incdir in incdirs.get_incdirs():
                 scan_command += ['-I%s' % os.path.join(state.environment.get_source_dir(), incdir)]
 
         if 'link_with' in kwargs:
             link_with = kwargs.pop('link_with')
+            if not isinstance(link_with, list):
+                link_with = [link_with]
             for link in link_with:
                 lib = link.held_object
                 scan_command += ['-l%s' % lib.name]
@@ -121,10 +137,13 @@ class GnomeModule:
         inc_dirs = None
         if kwargs.get('include_directories'):
             inc_dirs = kwargs.pop('include_directories')
-            if isinstance(inc_dirs.held_object, build.IncludeDirs):
-                scan_command += ['--add-include-path=%s' % inc for inc in inc_dirs.held_object.get_incdirs()]
-            else:
-                raise MesonException('Gir include dirs should be include_directories()')
+            if not isinstance(inc_dirs, list):
+                inc_dirs = [inc_dirs]
+            for id in inc_dirs:
+                if isinstance(id.held_object, build.IncludeDirs):
+                    scan_command += ['--add-include-path=%s' % inc for inc in id.held_object.get_incdirs()]
+                else:
+                    raise MesonException('Gir include dirs should be include_directories()')
         if isinstance(girtarget, build.Executable):
             scan_command += ['--program', girtarget]
         elif isinstance(girtarget, build.SharedLibrary):
@@ -144,8 +163,9 @@ class GnomeModule:
         typelib_output = '%s-%s.typelib' % (ns, nsversion)
         typelib_cmd = ['g-ir-compiler', scan_target, '--output', '@OUTPUT@']
         if inc_dirs:
-            typelib_cmd += ['--includedir=%s' % inc for inc in
-                    inc_dirs.held_object.get_incdirs()]
+            for id in inc_dirs:
+                typelib_cmd += ['--includedir=%s' % inc for inc in
+                                id.held_object.get_incdirs()]
         if deps:
             for dep in deps:
                 girdir = dep.held_object.get_variable ("girdir")
@@ -176,6 +196,48 @@ class GnomeModule:
         target_g = build.CustomTarget(targetname, state.subdir, kwargs)
         return target_g
 
+    def gtkdoc(self, state, args, kwargs):
+        if len(args) != 1:
+            raise MesonException('Gtkdoc must have one positional argument.')
+        modulename = args[0]
+        if not isinstance(modulename, str):
+            raise MesonException('Gtkdoc arg must be string.')
+        if not 'src_dir' in kwargs:
+            raise MesonException('Keyword argument src_dir missing.')
+        main_file = kwargs.get('main_sgml', '')
+        if not isinstance(main_file, str):
+            raise MesonException('Main sgml keyword argument must be a string.')
+        main_xml = kwargs.get('main_xml', '')
+        if not isinstance(main_xml, str):
+            raise MesonException('Main xml keyword argument must be a string.')
+        if main_xml != '':
+            if main_file != '':
+                raise MesonException('You can only specify main_xml or main_sgml, not both.')
+            main_file = main_xml
+        src_dir = kwargs['src_dir']
+        targetname = modulename + '-doc'
+        command = os.path.normpath(os.path.join(os.path.split(__file__)[0], "../gtkdochelper.py"))
+        if hasattr(src_dir, 'held_object'):
+            src_dir= src_dir.held_object
+            if not isinstance(src_dir, build.IncludeDirs):
+                raise MesonException('Invalidt keyword argument for src_dir.')
+            incdirs = src_dir.get_incdirs()
+            if len(incdirs) != 1:
+                raise MesonException('Argument src_dir has more than one directory specified.')
+            header_dir = os.path.join(state.environment.get_source_dir(), src_dir.get_curdir(), incdirs[0])
+        else:
+            header_dir = os.path.normpath(os.path.join(state.subdir, src_dir))
+        args = [state.environment.get_source_dir(),
+                state.environment.get_build_dir(),
+                state.subdir,
+                header_dir,
+                main_file,
+                modulename]
+        res = [build.RunTarget(targetname, command, args, state.subdir)]
+        if kwargs.get('install', True):
+            res.append(build.InstallScript([command] + args))
+        return res
+
     def gdbus_codegen(self, state, args, kwargs):
         if len(args) != 2:
             raise MesonException('Gdbus_codegen takes two arguments, name and xml file.')
