@@ -77,8 +77,6 @@ class NinjaBuildElement():
         self.elems.append((name, elems))
 
     def write(self, outfile):
-        # Sort inputs so the command line does not change.
-        # This allows Ninja to skip unnecessary rebuilds.
         line = 'build %s: %s %s' % (' '.join([ninja_quote(i) for i in self.outfilenames]),\
                                     self.rule,
                                     ' '.join([ninja_quote(i) for i in self.infilenames]))
@@ -116,6 +114,13 @@ class NinjaBackend(backends.Backend):
         self.source_suffix_in_objs = True
         self.ninja_filename = 'build.ninja'
         self.fortran_deps = {}
+        self.all_outputs = {}
+
+    def check_outputs(self, elem):
+        for n in elem.outfilenames:
+            if n in self.all_outputs:
+                raise MesonException('Multiple producers for Ninja target "%s". Please rename your targets.' % n)
+            self.all_outputs[n] = True
 
     def generate(self):
         outfilename = os.path.join(self.environment.get_build_dir(), self.ninja_filename)
@@ -164,7 +169,7 @@ class NinjaBackend(backends.Backend):
             self.generate_custom_target(target, outfile)
         if isinstance(target, build.RunTarget):
             self.generate_run_target(target, outfile)
-        name = target.get_basename() + target.type_suffix()
+        name = target.get_id()
         gen_src_deps = []
         if name in self.processed_targets:
             return
@@ -259,9 +264,9 @@ class NinjaBackend(backends.Backend):
             for src in self.generate_unity_files(target, unity_src):
                 obj_list.append(self.generate_single_compile(target, outfile, src, True, unity_deps + header_deps))
         linker = self.determine_linker(target, src_list)
-        # Sort object list to preserve command line over multiple invocations.
-        elem = self.generate_link(target, outfile, outname, sorted(obj_list), linker, pch_objects)
-        self.generate_shlib_aliases(target, self.get_target_dir(target), outfile, elem)
+        elem = self.generate_link(target, outfile, outname, obj_list, linker, pch_objects)
+        self.generate_shlib_aliases(target, self.get_target_dir(target))
+        elem.write(outfile)
         self.processed_targets[name] = True
 
     def process_target_dependencies(self, target, outfile):
@@ -279,7 +284,12 @@ class NinjaBackend(backends.Backend):
         ofilenames = [os.path.join(target.subdir, i) for i in target.output]
         # FIXME, should not grab element at zero but rather expand all.
         deps = [os.path.join(i.get_subdir(), self.hackety_hack(i.get_filename())) for i in target.get_dependencies()]
-        srcs = [os.path.join(self.build_to_src, target.subdir, i) for i in target.sources]
+        srcs = []
+        for i in target.sources:
+            if isinstance(i, str):
+                srcs.append(os.path.join(self.build_to_src, target.subdir, i))
+            else:
+                srcs.append(i.rel_to_builddir(self.build_to_src))
         deps +=  srcs
         if target.build_always:
             deps.append('PHONY')
@@ -305,17 +315,42 @@ class NinjaBackend(backends.Backend):
         elem.add_item('COMMAND', cmd)
         elem.add_item('description', 'Generating %s with a custom command.' % target.name)
         elem.write(outfile)
+        self.check_outputs(elem)
         self.processed_targets[target.name + target.type_suffix()] = True
 
     def generate_run_target(self, target, outfile):
         runnerscript = os.path.join(self.environment.get_script_dir(), 'commandrunner.py')
-        elem = NinjaBuildElement(target.name, 'CUSTOM_COMMAND', [])
-        cmd = [sys.executable, runnerscript, self.environment.get_source_dir(), self.environment.get_build_dir(),
-               target.subdir, target.command] + target.args
+        deps = []
+        arg_strings = []
+        for i in target.args:
+            if isinstance(i, str):
+                arg_strings.append(i)
+            elif isinstance(i, build.BuildTarget):
+                deps.append(self.get_target_filename(i))
+            else:
+                raise MesonException('Unreachable code.')
+        elem = NinjaBuildElement(target.name, 'CUSTOM_COMMAND', deps)
+        cmd = [sys.executable, runnerscript, self.environment.get_source_dir(), self.environment.get_build_dir(), target.subdir]
+        texe = target.command
+        try:
+            texe = texe.held_object
+        except AttributeError:
+            pass
+        if isinstance(texe, build.Executable):
+            abs_exe = os.path.join(self.environment.get_build_dir(), self.get_target_filename(texe))
+            deps.append(self.get_target_filename(texe))
+            if self.environment.is_cross_build() \
+                and self.environment.cross_info.config['binaries'].get('exe_wrapper', None) is not None:
+                cmd += [self.environment.cross_info.config['binaries']['exe_wrapper']]
+            cmd.append(abs_exe)
+        else:
+            cmd.append(target.command)
+        cmd += target.args
         elem.add_item('COMMAND', cmd)
         elem.add_item('description', 'Running external command %s.' % target.name)
         elem.add_item('pool', 'console')
         elem.write(outfile)
+        self.check_outputs(elem)
         self.processed_targets[target.name + target.type_suffix()] = True
 
     def generate_po(self, outfile):
@@ -328,6 +363,7 @@ class NinjaBackend(backends.Backend):
             elem.add_item('FILELIST', os.path.join(self.environment.get_source_dir(), input_file))
             elem.add_item('OUTDIR', os.path.join(self.environment.get_source_dir(), subdir))
             elem.write(outfile)
+            self.check_outputs(elem)
             for l in languages:
                 infile = os.path.join(self.environment.get_source_dir(), subdir, l + '.po')
                 outfilename = os.path.join(subdir, l + '.gmo')
@@ -335,6 +371,7 @@ class NinjaBackend(backends.Backend):
                 lelem.add_item('INFILE', infile)
                 lelem.add_item('OUTFILE', outfilename)
                 lelem.write(outfile)
+                self.check_outputs(lelem)
 
     def generate_coverage_rules(self, outfile):
         (gcovr_exe, lcov_exe, genhtml_exe) = environment.find_coverage_tools()
@@ -351,6 +388,7 @@ class NinjaBackend(backends.Backend):
                                       '-o', os.path.join(self.environment.get_log_dir(), 'coverage.txt')])
             elem.add_item('DESC', 'Generating text coverage report.')
             elem.write(outfile)
+            self.check_outputs(elem)
         if lcov_exe and genhtml_exe:
             added_rule = True
             phony_elem = NinjaBuildElement('coverage-html', 'phony', 'coveragereport/index.html')
@@ -364,6 +402,7 @@ class NinjaBackend(backends.Backend):
                        '--legend', '--show-details', 'coverage.info']
             elem.add_item('COMMAND', command)
             elem.add_item('DESC', 'Generating HTML coverage report.')
+            self.check_outputs(elem)
             elem.write(outfile)
         if not added_rule:
             mlog.log(mlog.red('Warning:'), 'coverage requested but neither gcovr nor lcov/genhtml found.')
@@ -381,6 +420,7 @@ class NinjaBackend(backends.Backend):
         elem.add_item('DESC', 'Installing files.')
         elem.add_item('COMMAND', [sys.executable, install_script, install_data_file])
         elem.add_item('pool', 'console')
+        self.generate_depmf_install(d)
         self.generate_target_install(d)
         self.generate_header_install(d)
         self.generate_man_install(d)
@@ -390,6 +430,7 @@ class NinjaBackend(backends.Backend):
         self.generate_custom_install_script(d)
         self.generate_subdir_install(d)
         elem.write(outfile)
+        self.check_outputs(elem)
 
         ofile = open(install_data_file, 'wb')
         pickle.dump(d, ofile)
@@ -434,8 +475,7 @@ class NinjaBackend(backends.Backend):
             d.man.append(i)
 
     def generate_custom_install_script(self, d):
-        d.install_script = self.build.install_script
-
+        d.install_scripts = self.build.install_scripts
 
     def generate_header_install(self, d):
         incroot = self.environment.get_includedir()
@@ -497,6 +537,7 @@ class NinjaBackend(backends.Backend):
         elem.add_item('DESC', 'Running test suite.')
         elem.add_item('pool', 'console')
         elem.write(outfile)
+        self.check_outputs(elem)
 
         if valgrind:
             velem = NinjaBuildElement('test-valgrind', 'CUSTOM_COMMAND', ['all', 'PHONY'])
@@ -504,6 +545,7 @@ class NinjaBackend(backends.Backend):
             velem.add_item('DESC', 'Running test suite under Valgrind.')
             velem.add_item('pool', 'console')
             velem.write(outfile)
+            self.check_outputs(velem)
 
     def generate_rules(self, outfile):
         outfile.write('# Rules for compiling.\n\n')
@@ -580,6 +622,7 @@ class NinjaBackend(backends.Backend):
         elem.add_dep([os.path.join(self.get_target_private_dir(target), i) for i in class_list])
         elem.add_item('ARGS', commands)
         elem.write(outfile)
+        self.check_outputs(elem)
 
     def generate_cs_resource_tasks(self, target, outfile):
         args = []
@@ -595,6 +638,7 @@ class NinjaBackend(backends.Backend):
                 elem.add_item('COMMAND', ['resgen', rel_sourcefile, ofilename])
                 elem.add_item('DESC', 'Compiling resource %s.' % rel_sourcefile)
                 elem.write(outfile)
+                self.check_outputs(elem)
                 deps.append(ofilename)
                 a = '-resource:' + ofilename
             else:
@@ -634,6 +678,7 @@ class NinjaBackend(backends.Backend):
         elem = NinjaBuildElement(outputs, 'cs_COMPILER', rel_srcs)
         elem.add_dep(deps)
         elem.add_item('ARGS', commands)
+        self.check_outputs(elem)
         elem.write(outfile)
 
     def generate_single_java_compile(self, subdir, src, target, compiler, outfile):
@@ -646,6 +691,7 @@ class NinjaBackend(backends.Backend):
         element = NinjaBuildElement(rel_obj, compiler.get_language() + '_COMPILER', rel_src)
         element.add_item('ARGS', args)
         element.write(outfile)
+        self.check_outputs(element)
         return plain_class_path
 
     def generate_java_link(self, outfile):
@@ -669,6 +715,7 @@ class NinjaBackend(backends.Backend):
             element = NinjaBuildElement(rel_vapi, valac.get_language() + '_COMPILER', rel_s)
             element.add_item('ARGS', args)
             element.write(outfile)
+            self.check_outputs(element)
             fastvapis[s] = (vapibase, rel_vapi)
         return fastvapis
 
@@ -731,6 +778,7 @@ class NinjaBackend(backends.Backend):
             element.add_orderdep(vapi_order_deps)
             element.add_dep(extra_dep_files)
             element.write(outfile)
+            self.check_outputs(element)
         return generated_c
 
     def generate_rust_target(self, target, outfile):
@@ -770,13 +818,17 @@ class NinjaBackend(backends.Backend):
         element.add_item('targetdep', depfile)
         element.add_item('cratetype', cratetype)
         element.write(outfile)
+        self.check_outputs(element)
 
     def generate_static_link_rules(self, is_cross, outfile):
         if self.build.has_language('java'):
             if not is_cross:
                 self.generate_java_link(outfile)
         if is_cross:
-            static_linker = self.build.static_cross_linker
+            if self.environment.cross_info.need_cross_compiler():
+                static_linker = self.build.static_cross_linker
+            else:
+                static_linker = self.build.static_linker
             crstr = '_CROSS'
         else:
             static_linker = self.build.static_linker
@@ -793,7 +845,15 @@ class NinjaBackend(backends.Backend):
         outfile.write(description)
 
     def generate_dynamic_link_rules(self, outfile):
-        ctypes = [(self.build.compilers, False), (self.build.cross_compilers, True)]
+        ctypes = [(self.build.compilers, False)]
+        if self.environment.is_cross_build():
+            if self.environment.cross_info.need_cross_compiler():
+                ctypes.append((self.build.cross_compilers, True))
+            else:
+                # Native compiler masquerades as the cross compiler.
+                ctypes.append((self.build.compilers, True))
+        else:
+            ctypes.append((self.build.cross_compilers, True))
         for (complist, is_cross) in ctypes:
             for compiler in complist:
                 langname = compiler.get_language()
@@ -801,11 +861,17 @@ class NinjaBackend(backends.Backend):
                  langname == 'rust' or langname == 'cs':
                     continue
                 crstr = ''
+                cross_args = []
                 if is_cross:
                     crstr = '_CROSS'
+                    try:
+                        cross_args = self.environment.cross_info.config['properties'][langname + '_link_args']
+                    except KeyError:
+                        pass
                 rule = 'rule %s%s_LINKER\n' % (langname, crstr)
-                command = ' command = %s $ARGS  %s $in $LINK_ARGS $aliasing\n' % \
+                command = ' command = %s %s $ARGS  %s $in $LINK_ARGS $aliasing\n' % \
                 (' '.join(compiler.get_linker_exelist()),\
+                 ' '.join(cross_args),\
                  ' '.join(compiler.get_linker_output_args('$out')))
                 description = ' description = Linking target $out'
                 outfile.write(rule)
@@ -926,8 +992,15 @@ rule FORTRAN_DEP_HACK
             if d != '$out' and d != '$in':
                 d = qstr % d
             quoted_depargs.append(d)
-        command = " command = %s $ARGS %s %s %s $in\n" % \
+        cross_args = []
+        if is_cross:
+            try:
+                cross_args = self.environment.cross_info.config['properties'][langname + '_args']
+            except KeyError:
+                pass
+        command = " command = %s %s $ARGS %s %s %s $in\n" % \
             (' '.join(compiler.get_exelist()),\
+             ' '.join(cross_args),
              ' '.join(quoted_depargs),\
              ' '.join(compiler.get_output_args('$out')),\
              ' '.join(compiler.get_compile_only_args()))
@@ -952,6 +1025,13 @@ rule FORTRAN_DEP_HACK
             crstr = ''
         rule = 'rule %s%s_PCH\n' % (langname, crstr)
         depargs = compiler.get_dependency_gen_args('$out', '$DEPFILE')
+        cross_args = []
+        if is_cross:
+            try:
+                cross_args = self.environment.cross_info.config['properties'][langname + '_args']
+            except KeyError:
+                pass
+
         quoted_depargs = []
         for d in depargs:
             if d != '$out' and d != '$in':
@@ -961,8 +1041,9 @@ rule FORTRAN_DEP_HACK
             output = ''
         else:
             output = ' '.join(compiler.get_output_args('$out'))
-        command = " command = %s $ARGS %s %s %s $in\n" % \
+        command = " command = %s %s $ARGS %s %s %s $in\n" % \
             (' '.join(compiler.get_exelist()),\
+             ' '.join(cross_args),\
              ' '.join(quoted_depargs),\
              output,\
              ' '.join(compiler.get_compile_only_args()))
@@ -985,7 +1066,13 @@ rule FORTRAN_DEP_HACK
             self.generate_compile_rule_for(langname, compiler, qstr, False, outfile)
             self.generate_pch_rule_for(langname, compiler, qstr, False, outfile)
         if self.environment.is_cross_build():
-            for compiler in self.build.cross_compilers:
+            # In case we are going a target-only build, make the native compilers
+            # masquerade as cross compilers.
+            if self.environment.cross_info.need_cross_compiler():
+                cclist = self.build.cross_compilers
+            else:
+                cclist = self.build.compilers
+            for compiler in cclist:
                 langname = compiler.get_language()
                 self.generate_compile_rule_for(langname, compiler, qstr, True, outfile)
                 self.generate_pch_rule_for(langname, compiler, qstr, True, outfile)
@@ -1031,13 +1118,14 @@ rule FORTRAN_DEP_HACK
                 else:
                     sole_output = ''
                 curfile = infilelist[i]
-                infilename = os.path.join(self.environment.get_source_dir(), curfile)
+                infilename = os.path.join(self.build_to_src, curfile)
                 outfiles = genlist.get_outputs_for(curfile)
                 outfiles = [os.path.join(self.get_target_private_dir(target), of) for of in outfiles]
                 args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)\
                         for x in base_args]
                 args = self.replace_outputs(args, self.get_target_private_dir(target), outfilelist)
-                args = [x.replace("@SOURCE_DIR@", self.environment.get_source_dir()).replace("@BUILD_DIR@", self.get_target_private_dir(target))
+                relout = os.path.join(target.subdir, target.get_basename() + target.type_suffix())
+                args = [x.replace("@SOURCE_DIR@", self.build_to_src).replace("@BUILD_DIR@", relout)
                         for x in args]
                 cmdlist = exe_arr + args
                 elem = NinjaBuildElement(outfiles, 'CUSTOM_COMMAND', infilename)
@@ -1048,6 +1136,7 @@ rule FORTRAN_DEP_HACK
                     elem.add_dep(self.get_target_filename(exe))
                 elem.add_item('COMMAND', cmdlist)
                 elem.write(outfile)
+                self.check_outputs(elem)
 
     def scan_fortran_module_outputs(self, target):
         compiler = None
@@ -1165,9 +1254,9 @@ rule FORTRAN_DEP_HACK
             basedir = i.get_curdir()
             for d in i.get_incdirs():
                 expdir =  os.path.join(basedir, d)
-                fulldir = os.path.join(self.environment.get_source_dir(), expdir)
+                srctreedir = os.path.join(self.build_to_src, expdir)
                 bargs = compiler.get_include_args(expdir)
-                sargs = compiler.get_include_args(fulldir)
+                sargs = compiler.get_include_args(srctreedir)
                 commands += bargs
                 commands += sargs
         if self.environment.coredata.use_pch:
@@ -1187,6 +1276,7 @@ rule FORTRAN_DEP_HACK
                 if srcfile == src:
                     depelem = NinjaBuildElement(modfile, 'FORTRAN_DEP_HACK', rel_obj)
                     depelem.write(outfile)
+                    self.check_outputs(depelem)
             commands += compiler.get_module_outdir_args(self.get_target_private_dir(target))
 
         element = NinjaBuildElement(rel_obj, compiler_name, rel_src)
@@ -1206,11 +1296,25 @@ rule FORTRAN_DEP_HACK
             element.add_orderdep(d)
         element.add_orderdep(pch_dep)
         element.add_orderdep(extra_orderdeps)
+        for i in self.get_fortran_orderdeps(target, compiler):
+            element.add_orderdep(i)
         element.add_item('DEPFILE', dep_file)
         element.add_item('ARGS', commands)
         element.write(outfile)
+        self.check_outputs(element)
         return rel_obj
 
+    # Fortran is a bit weird (again). When you link against a library, just compiling a source file
+    # requires the mod files that are output when single files are built. To do this right we would need to
+    # scan all inputs and write out explicit deps for each file. That is stoo slow and too much effort so
+    # instead just have an ordered dependendy on the library. This ensures all reqquired mod files are created.
+    # The real deps are then detected via dep file generation from the compiler. This breaks on compilers that
+    # produce incorrect dep files but such is life.
+    def get_fortran_orderdeps(self, target, compiler):
+        if compiler.language != 'fortran':
+            return []
+        return [os.path.join(lt.subdir, lt.filename) for lt in target.link_targets]
+
     def generate_msvc_pch_command(self, target, compiler, pch):
         if len(pch) != 2:
             raise RuntimeError('MSVC requires one header and one source to produce precompiled headers.')
@@ -1263,6 +1367,7 @@ rule FORTRAN_DEP_HACK
             elem.add_item('ARGS', commands)
             elem.add_item('DEPFILE', dep)
             elem.write(outfile)
+            self.check_outputs(elem)
         return pch_objects
 
     def generate_shsym(self, outfile, target):
@@ -1270,9 +1375,10 @@ rule FORTRAN_DEP_HACK
         targetdir = self.get_target_private_dir(target)
         symname = os.path.join(targetdir, target_name + '.symbols')
         elem = NinjaBuildElement(symname, 'SHSYM', target_name)
-        if self.environment.is_cross_build():
-            elem.add_item('CROSS', '--cross-host=' + self.environment.cross_info['name'])
+        if self.environment.is_cross_build() and self.environment.cross_info.need_cross_compiler():
+            elem.add_item('CROSS', '--cross-host=' + self.environment.cross_info.config['host_machine']['system'])
         elem.write(outfile)
+        self.check_outputs(elem)
 
     def generate_link(self, target, outfile, outname, obj_list, linker, extra_args=[]):
         if isinstance(target, build.StaticLibrary):
@@ -1336,6 +1442,7 @@ rule FORTRAN_DEP_HACK
         elem = NinjaBuildElement(outname, linker_rule, obj_list)
         elem.add_dep(dep_targets)
         elem.add_item('LINK_ARGS', commands)
+        self.check_outputs(elem)
         return elem
 
     def get_dependency_filename(self, t):
@@ -1343,19 +1450,19 @@ rule FORTRAN_DEP_HACK
             return os.path.join(self.get_target_private_dir(t), self.get_target_filename(t) + '.symbols')
         return self.get_target_filename(t)
 
-    def generate_shlib_aliases(self, target, outdir, outfile, elem):
+    def generate_shlib_aliases(self, target, outdir):
         basename = target.get_filename()
         aliases = target.get_aliaslist()
-        aliascmd = []
-        if shutil.which('ln'):
+        if not mesonlib.is_windows():
             for alias in aliases:
-                aliasfile = os.path.join(outdir, alias)
-                cmd = ["&&", 'ln', '-s', '-f', basename, aliasfile]
-                aliascmd += cmd
+                aliasfile = os.path.join(self.environment.get_build_dir(), outdir, alias)
+                try:
+                    os.remove(aliasfile)
+                except Exception:
+                    pass
+                os.symlink(basename, aliasfile)
         else:
-            mlog.log("Library versioning disabled because host does not support symlinks.")
-        elem.add_item('aliasing', aliascmd)
-        elem.write(outfile)
+            mlog.debug("Library versioning disabled because host does not support symlinks.")
 
     def generate_gcov_clean(self, outfile):
             gcno_elem = NinjaBuildElement('clean-gcno', 'CUSTOM_COMMAND', 'PHONY')
@@ -1364,6 +1471,7 @@ rule FORTRAN_DEP_HACK
             gcno_elem.add_item('COMMAND', [sys.executable, clean_script, '.', 'gcno'])
             gcno_elem.add_item('description', 'Deleting gcno files')
             gcno_elem.write(outfile)
+            self.check_outputs(elem, gcno_elem)
 
             gcda_elem = NinjaBuildElement('clean-gcda', 'CUSTOM_COMMAND', 'PHONY')
             script_root = self.environment.get_script_dir()
@@ -1371,6 +1479,7 @@ rule FORTRAN_DEP_HACK
             gcda_elem.add_item('COMMAND', [sys.executable, clean_script, '.', 'gcda'])
             gcda_elem.add_item('description', 'Deleting gcda files')
             gcda_elem.write(outfile)
+            self.check_outputs(gcda_elem)
 
     def is_compilable_file(self, filename):
         if filename.endswith('.cpp') or\
@@ -1396,6 +1505,7 @@ rule FORTRAN_DEP_HACK
                 infilename = os.path.join(self.build_to_src, target.get_source_subdir(), src)
                 elem = NinjaBuildElement(outfilename, rule.name, infilename)
                 elem.write(outfile)
+                self.check_outputs(elem)
                 if self.is_compilable_file(outfilename):
                     src_deps.append(outfilename)
                 else:
@@ -1408,6 +1518,7 @@ rule FORTRAN_DEP_HACK
 
         elem = NinjaBuildElement('all', 'phony', targetlist)
         elem.write(outfile)
+        self.check_outputs(elem)
 
         default = 'default all\n\n'
         outfile.write(default)
@@ -1423,6 +1534,7 @@ rule FORTRAN_DEP_HACK
             elem.add_dep('clean-gcda')
             elem.add_dep('clean-gcno')
         elem.write(outfile)
+        self.check_outputs(elem)
 
         deps = [os.path.join(self.build_to_src, df) \
                 for df in self.interpreter.get_build_def_files()]
@@ -1442,3 +1554,4 @@ rule FORTRAN_DEP_HACK
 
         elem = NinjaBuildElement(deps, 'phony', '')
         elem.write(outfile)
+        self.check_outputs(elem)
